/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Entry
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Santhosh N

#include "entry.h"
#include "font.h"
#include "image.h"
#include "eventloop.h"
#include "display.h"
#include "theme.h"
#include "widget_factory.h"

#include <X11/extensions/shape.h>
#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xutil.h>

#define BLINK_TIME	300000

#include <stdlib.h>
#include <string.h>

static void cursor_blink_timer(void*, void*);

static int bksp_fg=0, blink_fg=0, cursor_x=0;
static int select_fg=0, select_x=0, select_pos=0, select_len=0, motion_fg=0;
static px_timer *cursor_blink=NULL;

px_widget* entry_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();

	// Initialize common widget variables
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 
							0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	// This is needed to get SelectionClear events to the right window. Otherwise all
	// Clear events will be directed to the toplevel and we cannot make out when 
	// selection changes happen within the same toplevel.
	
	XSelectInput(display, widget->win, PropertyChangeMask);
	
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;

	if(tag != NULL) {
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;

	widget->wtype = ENTRY;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

	return widget;
}

void entry_set_parameters(px_widget* widget)
{
	px_entry *entry = &(widget->specific.entry);
	
	// Initialize common widget defaul callbacks
	widget->default_callbacks[SELECTION_CLEAR] = entry_default_selection_clear;
	widget->default_callbacks[BUTTON_PRESSED] = entry_default_button_pressed;
	widget->default_callbacks[BUTTON_RELEASED] = entry_default_button_released;
	widget->default_callbacks[MOTION] = entry_default_motion;
	widget->default_callbacks[KEY_PRESSED] = entry_default_key_pressed;
	widget->default_callbacks[KEY_RELEASED] = entry_default_key_released;

	widget->redraw = entry_redraw;
	widget->destroy = entry_destroy;
	
	entry->text = (char*)malloc(sizeof(char)*10);
	entry->text[0] = '\0';
	entry->allocated = 10;
	entry->fontname  = strdup("-picopeta-simputer-medium-r-normal--11-120-101-108-p-120-fontspecific-0");
	entry->font = font_load_font(entry->fontname);
	entry->show_from = 0;
	entry->cursor_position = 0;
	entry->password_char = 0;
	
	entry->foreground = (XColor*)malloc(sizeof(XColor));
	widget_xcolor_from_str(entry->foreground, "black");
	entry->disabled_foreground = (XColor*)malloc(sizeof(XColor));
	widget_xcolor_from_str(entry->disabled_foreground, "gray");

	entry->cut = entry_cut;
	entry->copy = entry_copy;
	entry->paste = entry_paste;
	
	entry->set_foreground_color = entry_set_foreground_color;
	entry->set_disabled_foreground_color = entry_set_disabled_color;
	
	entry->set_text = entry_set_text;
	entry->delete_text = entry_delete_text;
	entry->set_font = entry_set_font;
}

void entry_redraw(px_widget *widget)
{
	GC gc, mask_gc;
     px_entry *entry = &(widget->specific.entry);
     Display* display = display_get_Xdisplay();
	widget_states state;
	char *pstr = NULL;
	int text_width;

	state = (*widget->get_state)(widget);
     gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);

	// Draw the background
	if(widget->use_pixmap) {
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	} else {
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	// Create the gc for the mask
	if(widget->backing_store.mask != None) {
		mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		XSetFont(display, mask_gc, widget->specific.entry.font->font->fid);
		XSetForeground(display, mask_gc, 1);
	}

	// Set foreground color
	switch(state) {
		case DISABLED:
			if(widget->specific.entry.disabled_foreground)
				XSetForeground(display, gc, widget->specific.entry.disabled_foreground->pixel);
			break;
		case NORMAL:
			if(widget->specific.entry.foreground)
				XSetForeground(display, gc, widget->specific.entry.foreground->pixel);
			break;
		default:
			XSetForeground(display, gc, 0);
	}

	// Find the x-position of the cursor, for the whole string
	if(!entry->password_char) {
		cursor_x = font_get_width(entry->font, entry->text, entry->cursor_position);
	} else {
		pstr = (char*)malloc(strlen(entry->text)+10);
		memset(pstr, entry->password_char, strlen(entry->text));
		pstr[strlen(entry->text)] = 0;
		cursor_x = font_get_width(entry->font, pstr, entry->cursor_position);
	}

	// Adjust the view and cursor position
	if(cursor_x+entry->show_from > widget->width) {	// The cursor has reached right end
		entry->show_from = widget->width - cursor_x;
		cursor_x += entry->show_from-1;

		if(bksp_fg)
			bksp_fg = 0;
	} else if(cursor_x+entry->show_from<0) {		// The cursor has reached left end	
		if(!bksp_fg) {							// beacuse of left arrow key
			entry->show_from -= cursor_x+entry->show_from;
			cursor_x = 0;
		} else {								// because of backspace key
			if(-(entry->show_from) < widget->width) {
				entry->show_from = 0;
			} else {
				entry->show_from += widget->width;
			}
			
			cursor_x += entry->show_from-1;
			
			bksp_fg = 0;
		}
	} else {
		cursor_x += entry->show_from-1;
	
		if(cursor_x<0)
			cursor_x = 0;

		if(bksp_fg)
			bksp_fg = 0;
	}

	// Draw the text
	XSetFont(display, gc, entry->font->font->fid);
	if(!entry->password_char) {
		XDrawString(display, widget->backing_store.pxmp, gc, 
					entry->show_from, widget->height-entry->font->font->descent-4, 
					entry->text, strlen(entry->text));
	} else {
		XDrawString(display, widget->backing_store.pxmp, gc, 
					entry->show_from, widget->height-entry->font->font->descent-4, 
					pstr, strlen(pstr));
	}

	// Draw the bottom line
	XDrawLine(display, widget->backing_store.pxmp, gc, 0, widget->height-2, 
										widget->width, widget->height-2);

	// Highlight the selection OR Draw the cursor
	text_width = font_get_width(entry->font, entry->text, strlen(entry->text));
	if((select_x<text_width || cursor_x+1<text_width) && select_fg) {
		int i, slen, rx, rwidth;
		unsigned long bg_pixel, fg_pixel;

		slen = strlen(entry->text);

		select_pos = entry->cursor_position;
		
		if(select_x > cursor_x) {	// Forward selection
			for(i=0; i<slen-select_pos; i++)
				if(font_get_width(entry->font, entry->text+select_pos, i+1) > select_x-cursor_x)
					break;

			if(cursor_x==0)
				rx = cursor_x;
			else
				rx = cursor_x+1;
			rwidth = font_get_width(entry->font, entry->text+select_pos, i);
			select_len = i;
		} else {					// Backward selection
			while(select_pos>0) {
				select_pos--;
				if(font_get_width(entry->font, entry->text+select_pos, entry->cursor_position-select_pos) > cursor_x-select_x)
					break;
			}

			rwidth = font_get_width(entry->font, entry->text+select_pos, entry->cursor_position-select_pos);
			rx = cursor_x-rwidth+1;
			select_len = entry->cursor_position-select_pos;
		}
		
		bg_pixel = widget->specific.entry.foreground->pixel;
		fg_pixel = widget->background.background_color->first_color.pixel;

		XSetFunction(display, gc, GXxor);
		XSetForeground(display, gc, bg_pixel^fg_pixel);
		XFillRectangle(display, widget->backing_store.pxmp, gc, rx, 2, rwidth, widget->height-5);
		XSetFunction(display, gc, GXcopy);
	} else {
		if(!widget->have_focus)
			XSetForeground(display, gc, widget->background.background_color->first_color.pixel);
		
		XDrawLine(display, widget->backing_store.pxmp, gc, 
						cursor_x, widget->height - font_get_height(entry->font, "") - 5, 
						cursor_x, widget->height-4);

		select_fg = 0;
	}

	// Clean up
	if(pstr)
		free(pstr);

	XFreeGC(display, gc);
     if(widget->backing_store.mask != None)
     	XFreeGC(display, mask_gc);

	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);
}

void entry_destroy(px_widget* widget)
{
	px_eventloop* eloop = eventloop_get_eventloop();

	if(eloop->focussed_widget == widget)
     	eloop->focus(eloop, NULL);
		
	if(cursor_blink) {
		eloop->delete_timer(eloop, cursor_blink);
		cursor_blink = NULL;
	}
}

void entry_cut(px_widget* widget, const char* atom)
{
}

void entry_copy(px_widget* widget, const char* atom)
{
}

void entry_paste(px_widget* widget, const char* atom)
{
}

void entry_insert_text(px_widget* widget, const char* t, unsigned int index)
{
}

void entry_set_text(px_widget* widget, const char* t)
{
	px_entry *entry = &(widget->specific.entry);
	int slen, cursor_x;
	char *pstr=NULL;
	
	if(t==NULL)
		return;

	slen = strlen(t);

	// Set the text and allocate required memory
	if(entry->text)
		free(entry->text);
	entry->text = (char*)malloc(slen+10);
	entry->allocated = slen+10;
	strcpy(entry->text, t);
	
	entry->cursor_position = slen;
	
	// Get the cursor x-position
	if(!entry->password_char) {
		cursor_x = font_get_width(entry->font, entry->text, entry->cursor_position);
	} else {
		pstr = (char*)malloc(strlen(entry->text)+10);
		memset(pstr, entry->password_char, strlen(entry->text));
		pstr[strlen(entry->text)] = 0;
		cursor_x = font_get_width(entry->font, pstr, entry->cursor_position);
		free(pstr);
	}

	if(cursor_x>widget->width) {
		entry->show_from = widget->width - cursor_x;
	} else {
		entry->show_from = 0;
	}
}

void entry_delete_text(px_widget* widget, unsigned int del_from, unsigned int del_chars)
{
	px_entry *entry = &(widget->specific.entry);
	int slen, i;

	slen = strlen(entry->text);

	if(del_from >= slen)
		return;

	entry->show_from = 0;
	entry->cursor_position = 0;

	if(del_from+del_chars >= slen) {
		entry->text[del_from] = '\0';
	} else {
		for(i=del_from; i+del_chars<slen; i++)
			entry->text[i] = entry->text[i+del_chars];
		entry->text[i] = '\0';
	}
}

void entry_set_font(px_widget* widget, const char* fontname)
{
	if(widget->specific.entry.fontname) {
		font_free_font(widget->specific.entry.fontname);	
		widget->specific.entry.font = font_load_font(fontname);
		free(widget->specific.entry.fontname);
		widget->specific.entry.fontname = strdup(fontname);
	}
}

void entry_set_focus(px_widget* widget, Bool f)
{
	px_eventloop* eloop = eventloop_get_eventloop();
	
	if(cursor_blink) {
		eloop->delete_timer(eloop, cursor_blink);
		cursor_blink = NULL;
	}

	widget->have_focus = f;
	
	if(widget->have_focus) {
		cursor_blink = eloop->add_timer(eloop, 0, BLINK_TIME, cursor_blink_timer, widget, NULL, True);
	}
		
	widget->redraw(widget);
	widget->show(widget);
}

void entry_set_foreground_color(px_widget* widget, const char* colour)
{
	if(widget->specific.entry.foreground == NULL) {
		widget->specific.entry.foreground = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.entry.foreground, colour);
}

void entry_set_disabled_color(px_widget* widget, const char* colour)
{
	if(widget->specific.entry.disabled_foreground == NULL) {
		widget->specific.entry.disabled_foreground = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.entry.disabled_foreground, colour);
}

Bool entry_default_button_pressed(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
     px_entry *entry = &(widget->specific.entry);
	px_widget *prev_widget;
	int slen, elen, i;

	eventloop->grab(eventloop, widget);
	
	// Set the selection flag to 0
	select_fg = 0;

	motion_fg = 1;

	// Set the focus, if not set already
	if(eventloop->focussed_widget != widget) {
		prev_widget = eventloop->focussed_widget;
     	eventloop->focus(eventloop, widget);
		if(prev_widget)
			prev_widget->redraw(prev_widget);
	}
	
	// Change the cursor position
	elen = abs(entry->show_from) + event->x;

	if(!entry->password_char) {
		slen = strlen(entry->text);
		if(elen >= font_get_width(entry->font, entry->text, slen)) {
			entry->cursor_position = slen;
		} else {
			for(i=0; i<slen; i++) {
				if(font_get_width(entry->font, entry->text, i) > elen)
					break;
			}
			entry->cursor_position = i-1;
		}
	} else {
		char *pstr;

		slen = strlen(entry->text);
		pstr = (char*)malloc(slen+10);
		memset(pstr, entry->password_char, slen);
		pstr[slen] = '\0';
		
		if(elen >= font_get_width(entry->font, pstr, slen)) {
			entry->cursor_position = slen;
		} else {
			for(i=0; i<slen; i++) {
				if(font_get_width(entry->font, pstr, i) > elen)
					break;
			}
			entry->cursor_position = i-1;
		}

		free(pstr);
	}
	
	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool entry_default_button_released(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();

	motion_fg = 0;
	
	eventloop->grabbed_widget = NULL;

	return True;
}

Bool entry_default_motion(px_widget *widget, px_event *event)
{
	px_entry *entry = &(widget->specific.entry);
	Display* display = display_get_Xdisplay();
	int fg=0;
	XEvent xev, xev_copy;

	if(!motion_fg)
		return True;

	// Selection allowed only on text entries
	if(!entry->password_char) {
		select_fg = 1;
		select_x = event->x;
	}

	// Bypass extra motion events
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(!fg && xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg) {
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);
	}

	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool entry_default_key_pressed(px_widget* widget, px_event* event)
{
	px_entry *entry = &(widget->specific.entry);
	KeySym keysym;
	char buffer[10];
	int buffer_length = 0, i, slen;
	
	if(entry->text==NULL)
		return False;

	buffer_length = XLookupString(&event->xev->xkey, buffer, 10, &keysym, NULL);

	event->ch = buffer[0];

	slen = strlen(entry->text);

	switch(keysym)	{
		case XK_Left:
			if(entry->cursor_position) {
				entry->cursor_position--;
			}
			break;

		case XK_Right:
			if(entry->cursor_position < slen) {
				entry->cursor_position++;
			}
			break;

		case XK_Home:
			entry->cursor_position = 0;
			break;

		case XK_End:
			entry->cursor_position = slen;
			break;

		case XK_BackSpace:
			if(entry->cursor_position && !bksp_fg) {
				entry->cursor_position--;
				for(i=entry->cursor_position; i<slen; i++)
					entry->text[i] = entry->text[i+1];
				entry->text[slen-1] = '\0';
				bksp_fg = 1;
			}
			break;

		case XK_Delete:
			for(i=entry->cursor_position; i<slen; i++) {
				entry->text[i] = entry->text[i+1];
			}
			
			if(entry->cursor_position!=slen) {
				entry->text[slen-1] = '\0';
			}

			break;
		
		case XK_Return:
			break;

		default :
			if(keysym>=XK_space && keysym<=XK_asciitilde) {
				// Check if, memory reallocation needs to be done
				if(slen == entry->allocated-3) {
					char *tmp;
					tmp = realloc(entry->text, entry->allocated+10);
					if(tmp==NULL)
						return False;
					entry->text = tmp;
					entry->allocated += 10;
				}

				for(i=slen; i>entry->cursor_position; i--) {
					entry->text[i] = entry->text[i-1];
				}
				entry->text[slen+1] = '\0';
				entry->text[entry->cursor_position] = buffer[0];
				entry->cursor_position++;
			}
	}

	select_fg = 0;

	widget->redraw(widget);
	widget->show(widget);
	
	return True;
}

Bool entry_default_key_released(px_widget* widget, px_event* event)
{
	KeySym keysym;
	char buffer[10];
	
	XLookupString(&event->xev->xkey, buffer, 10, &keysym, NULL);

	event->ch = buffer[0];
	
	return True;
}

static void cursor_blink_timer(void *var1, void *var2)
{
	GC gc;
	px_widget *widget = (px_widget*)var1; 
	px_entry *entry = &(widget->specific.entry);
	Display* display = display_get_Xdisplay();
	widget_states state;

	state = (*widget->get_state)(widget);

	if(state==DISABLED || select_fg)
		return;

     gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);

	if(blink_fg) {
		XSetForeground(display, gc, widget->background.background_color->first_color.pixel);
		blink_fg = 0;
	} else {
		XSetForeground(display, gc, widget->specific.entry.foreground->pixel);
		blink_fg = 1;
	}

	XDrawLine(display, widget->backing_store.pxmp, gc, 
					cursor_x, widget->height - font_get_height(entry->font, "") - 5, 
					cursor_x, widget->height-4);
	XClearWindow(display, widget->win);

	XFreeGC(display, gc);
}

Bool entry_default_selection_clear(px_widget* widget, px_event* event)
{
	return True;
}
